//-----------------------------------------------------------------------------
// F50x_Comm_LIN0.c
//-----------------------------------------------------------------------------
// Copyright (C) 2010 Silicon Laboratories, Inc.
// http://www.silabs.com
//
// Program Description:
//
// [*** TEMPLATE INFORMATION ***]
// Filename: Fnnn_Comm_CommType.c
// nnn => Fill-in with the MCU family part number.
// CommType => Fill-in Comm interface name.
// Examples: F50x_Comm_CAN0.c, F34x_Comm_UART0.c
//
//
//
//
//
// Target:         C8051Fxxx (Any Silicon Labs Flash MCU)
// Tool chain:     Generic
// Command Line:   None
//
//
// Release 1.0 / 01SEP2010 (PKC)
//    -Initial Revision
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <si_toolchain.h>
#include "Fxxx_SFR_Defs_Wrapper.h"     // This header file will include the
                                       // real MCU register definition file

#include "Fxxx_Target_Config.h"
#include "Fxxx_Target_Interface.h"

#include "Fxxx_TargetBL_Config.h"
#include "Fxxx_TargetBL_Interface.h"

#include "Fxxx_BL130_LIN0_Interface.h"  // TODO: Change to actual file name - DONE

//-----------------------------------------------------------------------------
// Global CONSTANTS
//-----------------------------------------------------------------------------
#define RSP_OK 0x00

#define READ_LIN0_REG(addr, target) { LIN0ADR = addr; target = LIN0DAT; }
#define WRITE_LIN0_REG(addr1, data1) { LIN0ADR = addr1; LIN0DAT = data1; }

#if (SYSCLK < 8000000)                 // For the slowest baud rate
#error Minimum system clock needed to operate LIN peripheral is 8Mhz.
#endif

// Enabling this is useful for debugging
#define ENABLE_ERROR_COUNTERS 0 // 0=Disabled; 1=ENABLED

//-----------------------------------------------------------------------------
// LIN #defines
//-----------------------------------------------------------------------------

//------ Control register ------
#define  STREQ       0x01              // LIN Master Start transmission request
#define  WUPREQ      0x02              // Wake-Up request
#define  RSTERR      0x04              // Error reset request
#define  RSTINT      0x08              // Interrupt reset request
#define  DTACK       0x10              // Slave Data ACK
#define  TXRX        0x20              // Master selection of TX or RX mode
#define  SLEEP       0x40              // Slave flag set by master to inform
                                       // the peripheral that either a sleep
                                       // mode frame or bus idle timeout
                                       // was reached
#define  STOP        0x80              // Slave flag, to be set by the
                                       // application so no processing is to be
                                       // done until next SYNCH BREAK
//------ Status Register ------
#define  DONE        0x01              // Msg. processing completed
#define  WAKEUP      0x02              // Wake-Up signal
#define  ERROR       0x04              // Error detected
#define  LININTREQ   0x08              // Interrupt request
#define  DTREQ       0x10              // ID detected
#define  ABORT       0x20              // Abort requested
#define  IDLTOUT     0x40              // Time-out detected
#define  ACTIVE      0x80              // Interface active (communicating)

#if (AUTO_BAUD == 0)                   // If running in manual baud rate mode

 #define LIN_MULTIPLIER (uint8_t) ((20000.0 / LIN_BAUD_RATE) - 1)

 #if (SYSCLK >= 16000000)
   #define LIN_PRESCALAR 1
   #define LIN_DIVIDER (uint16_t) (SYSCLK / ((uint32_t) 4 * (LIN_MULTIPLIER + 1) * (uint32_t) LIN_BAUD_RATE))
 #else
   #define LIN_PRESCALAR 0
   #define LIN_DIVIDER (uint16_t) (SYSCLK / ((uint32_t) 2 * (LIN_MULTIPLIER + 1) * (uint32_t) LIN_BAUD_RATE))
 #endif

#else                                  // If running in auto baud rate mode

   #if (SYSCLK >= 16000000)
      #define LIN_PRESCALAR 1
      #define LIN_DIVIDER (uint16_t)((uint32_t) SYSCLK / ((uint32_t) 4 * (uint32_t) 20000))
   #else
      #define LIN_PRESCALAR 0
      #define LIN_DIVIDER (uint16_t)((uint32_t) SYSCLK / ((uint32_t) 2 * (uint32_t) 20000))
   #endif

#endif

//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------
uint8_t LIN_Error = 0;

//-----------------------------------------------------------------------------
// Function Prototypes (Local)
//-----------------------------------------------------------------------------
void Comm_Init (void);
uint8_t Comm_Wait_For_Rx_Bytes (uint16_t numbytes_to_read);
void Comm_Send_Tx_Bytes (uint16_t numbytes_to_send);

uint8_t Comm_Wait_For_Rx_RSP_OK (void);
void Comm_Send_Tx_RSP_OK (void);

//=============================================================================
// Function Definitions
//=============================================================================

//-----------------------------------------------------------------------------
// Comm_Init
//-----------------------------------------------------------------------------
//
// Return Value : None
// Parameters   : None
//
// Intializes the LIN peripheral to slave mode.
//
// LIN is an asynchronous, serial communications interface. All LIN registers
// besides the control mode register (LIN0CF) are indirectly accessed;
// the address is specified in the LINADDR register which points to a specific
// data register which can be accessed through the LINDATA register.
// Waveforms and additional specifications can be found in the LIN consortium
// (http:\\www.lin-subbus.com).
//
//-----------------------------------------------------------------------------
void Comm_Init (void)
{
   // *** Comm-Specific / Device-Specific ***
   // TODO: Add code here to initialize the Comm Interface

   //volatile uint16_t divider;

   uint8_t SFRPAGE_save = SFRPAGE;
   SFRPAGE = CONFIG_PAGE;
   // Enable interface, configure the device for LIN Slave, manual baud rate
   LIN0CF = 0x00;
   
   #if (AUTO_BAUD == 1)
   LIN0CF |= 0x20;                     // Automatic baud rate
   #endif

   LIN0CF |= 0x80;

   SFRPAGE = ACTIVE_PAGE;
   // Load the least significant byte of the divider.
   LIN0ADR = LIN0DIV;
   LIN0DAT = (uint8_t) LIN_DIVIDER;

   //divider = LIN_DIVIDER;

   LIN0ADR = LIN0MUL;

   #if (AUTO_BAUD == 0)
   // Load the multiplier, prescalar, and most significant bit of the divider.
   LIN0DAT = (uint8_t) ((LIN_PRESCALAR << 6) | (LIN_MULTIPLIER << 1)
               | (LIN_DIVIDER >> 8));
   #else
   // Load the prescalar and most significant bit of the divider
   LIN0DAT = (uint8_t) ((LIN_PRESCALAR << 6) | (LIN_DIVIDER >> 8));
   #endif

   LIN0ADR = LIN0SIZE;
   LIN0DAT = 0x80 | LIN_MESSAGE_SIZE;                     // Enhanced checksum & 8 data bytes

   //EIE1 |= 0x80;                       // Enable LIN interrupts
   // *** Do not enable LIN0 interrupts -- Polled-mode ***

   LIN0ADR = LIN0CTRL;
   LIN0DAT = 0x0C;                     // Reset LININT and LIN0ERR

   SFRPAGE = SFRPAGE_save;
}

//-----------------------------------------------------------------------------
// Comm_Wait_For_Rx_RSP_OK
//-----------------------------------------------------------------------------
//
// Return Value:  None
// Parameters:    None
//
//
//-----------------------------------------------------------------------------
uint8_t Comm_Wait_For_Rx_RSP_OK (void)
{
   // *** Comm-Specific / Device-Specific ***
   // TODO: Add code here to read bytes from the comm interface and store them
   // in the "Rx_Buf" buffer (buffer variable defined elsewhere).
   // Return the 1st byet of Rx_Buf as the return value.
   // This should be a blocking function; wait until requested number of bytes
   // have been read from the comm interface.
   // Send an ack (RSP_OK) after receiving each packet (except for the last packet)

   uint8_t status;
   uint8_t id;
   bit rx_complete_flag = 0;
   bit good_id = 0;
   uint8_t rsp;

   #if ENABLE_ERROR_COUNTERS == 1
   uint8_t error;
   static uint8_t num_error = 0;
   #endif

   uint8_t SFRPAGE_save = SFRPAGE;
   SFRPAGE  = ACTIVE_PAGE;

   while (rx_complete_flag == 0)
   {
      LIN0ADR = LIN0ST; // Poll the LIN0 status register

      do
      {
         status = LIN0DAT;             // Read status
      }
      while ((status & LININTREQ) == 0);    // Wait until interrupt bit is set

      if(status & DTREQ)               // If ID field received
      {
         READ_LIN0_REG(LIN0ID, id);

         if (id == ID_TARGET_RX_BL_CMD)
         {
            LIN0ADR = LIN0CTRL;
            LIN0DAT &= ~TXRX;          // Select to receive data

            LIN0ADR = LIN0SIZE;
            LIN0DAT = (LIN0DAT & 0xF0) | LIN_MESSAGE_SIZE;  // 8 data bytes

            LIN0ADR = LIN0CTRL;
            LIN0DAT |= DTACK;                   // Acknowledge data

            good_id = 1;
         }
         else
         {
            // Unexpected ID!

            good_id = 0;

            LIN0ADR = LIN0CTRL;
            LIN0DAT |= STOP;                   // Block further processing until next SYNC BREAK
         }
      }
      else if (status & ERROR)         // If an error was detected
      {
         #if ENABLE_ERROR_COUNTERS == 1
         READ_LIN0_REG(LIN0ERR, error);   // Check the error register
         // Keep track of the number of errors received
         num_error++;
         #endif

         // Handle your errors here, application dependent
      }
      else if(status & DONE)           // If reception completed
      {
         // Transmission complete

         if (good_id == 1)
         {
            READ_LIN0_REG(LIN0DT1, rsp);

            // Ignore other data bytes

            rx_complete_flag = 1;
            good_id = 0;
         }
      }

      LIN0ADR = LIN0CTRL;
      LIN0DAT |= 0x0C;                    // Reset interrupt request, error flag
   }

   SFRPAGE = SFRPAGE_save;
   
   return (rsp); // Return the response code
}

//-----------------------------------------------------------------------------
// Comm_Wait_For_Rx_Bytes
//-----------------------------------------------------------------------------
//
// Return Value:  None
// Parameters:    None
//
//
//-----------------------------------------------------------------------------
uint8_t Comm_Wait_For_Rx_Bytes (uint16_t numbytes_to_read)
{
   // *** Comm-Specific / Device-Specific ***
   // TODO: Add code here to read bytes from the comm interface and store them
   // in the "Rx_Buf" buffer (buffer variable defined elsewhere).
   // Return the 1st byet of Rx_Buf as the return value.
   // This should be a blocking function; wait until requested number of bytes
   // have been read from the comm interface.
   // Send an ack (RSP_OK) after receiving each packet (except for the last packet)

   uint8_t status;
   uint8_t id;
   uint8_t received_numbytes = 0;
   bit rx_complete_flag = 0;
   bit good_id = 0;
   uint8_t i;

   #if ENABLE_ERROR_COUNTERS == 1
   uint8_t error;
   static uint8_t num_error = 0;
   #endif  

   uint8_t SFRPAGE_save = SFRPAGE;
   SFRPAGE  = ACTIVE_PAGE;

   while (received_numbytes < numbytes_to_read)
   {
      while (rx_complete_flag == 0)
      {
         LIN0ADR = LIN0ST; // Poll the LIN0 status register

         do
         {
            status = LIN0DAT;             // Read status
         }
         while ((status & LININTREQ) == 0);    // Wait until interrupt bit is set

         if(status & DTREQ)               // If ID field received
         {
            READ_LIN0_REG(LIN0ID, id);

            if ( (id == ID_TARGET_RX_BL_CMD) || 
                 ( (Expect_WR_Command_Data_Bytes == 1) && (id == ID_TARGET_RX_WR_CMD) ) )
            {
               LIN0ADR = LIN0CTRL;
               LIN0DAT &= ~TXRX;          // Select to receive data

               LIN0ADR = LIN0SIZE;
               LIN0DAT = (LIN0DAT & 0xF0) | LIN_MESSAGE_SIZE;  // 8 data bytes

               LIN0ADR = LIN0CTRL;
               LIN0DAT |= DTACK;                   // Acknowledge data

               good_id = 1;
            }
            else
            {
               // Unexpected ID!

               good_id = 0;

               LIN0ADR = LIN0CTRL;
               LIN0DAT |= STOP;                   // Block further processing until next SYNC BREAK
            }
         }
         else if (status & ERROR)         // If an error was detected
         {

         #if ENABLE_ERROR_COUNTERS == 1
         READ_LIN0_REG(LIN0ERR, error);   // Check the error register
         // Keep track of the number of errors received
         num_error++;
         #endif

            // Handle your errors here, application dependent

         }
         else if(status & DONE)           // If reception completed
         {
            // Transmission complete

            if (good_id == 1)
            {
               // Read all 8 data bytes to *(Rx_Buf+], even though all bytes might not be valid
               for (i = 0; i < LIN_MESSAGE_SIZE; i++)
               {
                  READ_LIN0_REG((LIN0DT1 + i), *(Rx_Buf + i + received_numbytes));
               }
   
               received_numbytes += LIN_MESSAGE_SIZE;
               rx_complete_flag = 1;

               good_id = 0;
            }
         }

         LIN0ADR = LIN0CTRL;
         LIN0DAT |= 0x0C;                    // Reset interrupt request, error flag
      }

      rx_complete_flag = 0;

      if (received_numbytes < numbytes_to_read)
      {
         // For all packets except the last packet, 
         // send ack to master to get the next packet

         //*(Tx_Buf + 0) = RSP_OK;
         //Comm_Send_Tx_Bytes(1);

         Comm_Send_Tx_RSP_OK();
      }
   }
   
   SFRPAGE = SFRPAGE_save;
   
   return (*(Rx_Buf + 0)); // Return the BL command code, which is always at byte0
}

//-----------------------------------------------------------------------------
// Comm_Send_Tx_RSP_OK
//-----------------------------------------------------------------------------
//
// Return Value : None
// Parameters   : None
//
// Send the data in the buffer using the message object. 
//
//-----------------------------------------------------------------------------
void Comm_Send_Tx_RSP_OK (void)
{
   // *** Comm-Specific / Device-Specific ***
   // TODO: Add code here to write bytes to the comm interface from
   // the "Tx_Buf" buffer (buffer variable defined elsewhere).
   // This should be a blocking function; wait until ALL of the requested number 
   // of bytes have been sent via the comm interface.

   // Wait for an ack (RSP_OK) after sending each packet (except for the last packet)

   // This function assumes that the message object is fully initialized
   // in CAN0_Init and so all it has to do is fill the data registers and
   // initiate transmission
   
   uint8_t status;
   uint8_t id;
   bit tx_complete_flag = 0;
   bit tx_pending_flag = 0;

   #if ENABLE_ERROR_COUNTERS == 1
   uint8_t error;
   static uint8_t num_error = 0;
   #endif  
      
   uint8_t SFRPAGE_save = SFRPAGE;
   SFRPAGE  = ACTIVE_PAGE;

   while (tx_complete_flag == 0)
   {      
      LIN0ADR = LIN0ST; // Poll the LIN0 status register

      do
      {
         status = LIN0DAT;             // Read status
      }
      while ((status & LININTREQ) == 0);    // Wait until interrupt bit is set

      if(status & DTREQ)               // If ID field received
      {
         READ_LIN0_REG(LIN0ID, id);

         if (id == ID_TARGET_TX_BL_RSP)
         {
            LIN0ADR = LIN0CTRL;
            LIN0DAT |= TXRX;           // Select to transmit data

            LIN0ADR = LIN0SIZE;
            LIN0DAT = (LIN0DAT & 0xF0) | LIN_MESSAGE_SIZE;  //transmit 8 data bytes

            WRITE_LIN0_REG(LIN0DT1, RSP_OK);            // Initialize data registers

            // Other data bytes are "don't-care"
            
            LIN0ADR = LIN0CTRL;
            LIN0DAT |= DTACK;                   // Acknowledge data

            tx_pending_flag = 1;
         }
         else
         {
            // Unexpected ID!

            LIN0ADR = LIN0CTRL;
            LIN0DAT |= STOP;                   // Block further processing until next SYNC BREAK
         }
      }
      else if (status & ERROR)         // If an error was detected
      {

         #if ENABLE_ERROR_COUNTERS == 1
         READ_LIN0_REG(LIN0ERR, error);   // Check the error register
         // Keep track of the number of errors received
         num_error++;
         #endif

         // Handle your errors here, application dependent

      }
      else if(status & DONE)
      {
         // Transmission complete

         if (tx_pending_flag == 1)
         {
            tx_complete_flag = 1;
         }
      }

      LIN0ADR = LIN0CTRL;
      LIN0DAT |= 0x0C;                    // Reset interrupt request, error flag
   }
   
   SFRPAGE = SFRPAGE_save;             // Restore SFRPAGE
}


//-----------------------------------------------------------------------------
// Comm_Send_Tx_Bytes
//-----------------------------------------------------------------------------
//
// Return Value : None
// Parameters   : None
//
// Send the data in the buffer using the message object. 
//
//-----------------------------------------------------------------------------
void Comm_Send_Tx_Bytes (uint16_t numbytes_to_send)
{
   // *** Comm-Specific / Device-Specific ***
   // TODO: Add code here to write bytes to the comm interface from
   // the "Tx_Buf" buffer (buffer variable defined elsewhere).
   // This should be a blocking function; wait until ALL of the requested number 
   // of bytes have been sent via the comm interface.

   // Wait for an ack (RSP_OK) after sending each packet (except for the last packet)

   uint16_t sent_numbytes = 0;
   uint8_t status;
   uint8_t id;
   bit tx_complete_flag = 0;
   bit tx_pending_flag = 0;
   uint8_t i;

   #if ENABLE_ERROR_COUNTERS == 1
   uint8_t error;
   static uint8_t num_error = 0;
   #endif  
      
   uint8_t SFRPAGE_save = SFRPAGE;
   SFRPAGE  = ACTIVE_PAGE;

   while (sent_numbytes < numbytes_to_send)
   {
      while (tx_complete_flag == 0)
      {      
         LIN0ADR = LIN0ST; // Poll the LIN0 status register

         do
         {
            status = LIN0DAT;             // Read status
         }
         while ((status & LININTREQ) == 0);    // Wait until interrupt bit is set

         if(status & DTREQ)               // If ID field received
         {
            READ_LIN0_REG(LIN0ID, id);

            if (id == ID_TARGET_TX_BL_RSP)
            {
               LIN0ADR = LIN0CTRL;
               LIN0DAT |= TXRX;           // Select to transmit data

               LIN0ADR = LIN0SIZE;
               LIN0DAT = (LIN0DAT & 0xF0) | LIN_MESSAGE_SIZE;  //transmit 8 data bytes

               for (i = 0; i < LIN_MESSAGE_SIZE; i++)
               {
                  WRITE_LIN0_REG((LIN0DT1 + i), *(Tx_Buf + i + sent_numbytes));            // Initialize data registers
               }

               LIN0ADR = LIN0CTRL;
               LIN0DAT |= DTACK;                   // Acknowledge data
          
               tx_pending_flag = 1;
            }
            else
            {
               // Unexpected ID!

               LIN0ADR = LIN0CTRL;
               LIN0DAT |= STOP;                   // Block further processing until next SYNC BREAK
            }
         }
         else if (status & ERROR)         // If an error was detected
         {

            #if ENABLE_ERROR_COUNTERS == 1
            READ_LIN0_REG(LIN0ERR, error);   // Check the error register
            // Keep track of the number of errors received
            num_error++;
            #endif

            // Handle your errors here, application dependent

         }
         else if(status & DONE)
         {
            // Transmission complete

            if (tx_pending_flag == 1)
            {
               sent_numbytes += LIN_MESSAGE_SIZE;     
               tx_complete_flag = 1;
            }
         }

         LIN0ADR = LIN0CTRL;
         LIN0DAT |= 0x0C;                    // Reset interrupt request, error flag
      }

      tx_complete_flag = 0;

      if (sent_numbytes < numbytes_to_send)
      {
         // For all packets except the last packet, 
         // wait for ack from master before sending next packet

         //if (Comm_Wait_For_Rx_Bytes(1) != RSP_OK)
         if (Comm_Wait_For_Rx_RSP_OK() != RSP_OK)
         {
            // Error: Unexpected response
         }
      }

   }

   SFRPAGE = SFRPAGE_save;             // Restore SFRPAGE
}
//-----------------------------------------------------------------------------
// End Of File
//-----------------------------------------------------------------------------
